/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/indexes/bmaindex.hpp>
#include <ql/indexes/equityindex.hpp>
#include <ql/indexes/iborindex.hpp>
#include <ql/indexes/indexmanager.hpp>
#include <ql/indexes/inflationindex.hpp>
#include <ql/indexes/interestrateindex.hpp>
#include <ql/indexes/region.hpp>
#include <ql/indexes/swapindex.hpp>

#include <ql/indexes/ibor/all.hpp>
#include <ql/indexes/inflation/all.hpp>
#include <ql/indexes/swap/all.hpp>
