/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/models/shortrate/onefactormodel.hpp>
#include <ql/models/shortrate/twofactormodel.hpp>

#include <ql/models/shortrate/calibrationhelpers/all.hpp>
#include <ql/models/shortrate/onefactormodels/all.hpp>
#include <ql/models/shortrate/twofactormodels/all.hpp>
