/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/patterns/curiouslyrecurring.hpp>
#include <ql/patterns/lazyobject.hpp>
#include <ql/patterns/observable.hpp>
#include <ql/patterns/singleton.hpp>
#include <ql/patterns/visitor.hpp>

