/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/math/optimization/armijo.hpp>
#include <ql/math/optimization/bfgs.hpp>
#include <ql/math/optimization/conjugategradient.hpp>
#include <ql/math/optimization/constraint.hpp>
#include <ql/math/optimization/costfunction.hpp>
#include <ql/math/optimization/differentialevolution.hpp>
#include <ql/math/optimization/endcriteria.hpp>
#include <ql/math/optimization/goldstein.hpp>
#include <ql/math/optimization/leastsquare.hpp>
#include <ql/math/optimization/levenbergmarquardt.hpp>
#include <ql/math/optimization/linesearch.hpp>
#include <ql/math/optimization/linesearchbasedmethod.hpp>
#include <ql/math/optimization/lmdif.hpp>
#include <ql/math/optimization/method.hpp>
#include <ql/math/optimization/problem.hpp>
#include <ql/math/optimization/projectedconstraint.hpp>
#include <ql/math/optimization/projectedcostfunction.hpp>
#include <ql/math/optimization/projection.hpp>
#include <ql/math/optimization/simplex.hpp>
#include <ql/math/optimization/simulatedannealing.hpp>
#include <ql/math/optimization/spherecylinder.hpp>
#include <ql/math/optimization/steepestdescent.hpp>

